--[[
Copyright (C) GtX (Andy), 2020

Author: GtX | Andy
Date: 05.05.2020
Revision: FS22-01

Contact:
https://forum.giants-software.com
https://github.com/GtX-Andy

Important:
Not to be added to any mods / maps or modified from its current release form.
No modifications may be made to this script, including conversion to other game versions without written permission from GtX | Andy

Darf nicht zu Mods / Maps hinzugefügt oder von der aktuellen Release-Form geändert werden.
Ohne schriftliche Genehmigung von GtX | Andy dürfen keine Änderungen an diesem Skript vorgenommen werden, einschließlich der Konvertierung in andere Spielversionen
]]

MultiplayerVehicleKeysManager = {}

local validationFail
local popupMessage

local buildId = 1
local versionString = "0.0.0.0"

local modName = g_currentModName or ""
local modDirectory = g_currentModDirectory or ""

local MultiplayerVehicleKeysManager_mt = Class(MultiplayerVehicleKeysManager)

function MultiplayerVehicleKeysManager.new(customMt)
    local self = setmetatable({}, customMt or MultiplayerVehicleKeysManager_mt)

    self.isServer = g_server ~= nil
    self.isClient = g_client ~= nil

    self.numInserted = 0

    self.currentVehicleKeys = {}
    self.keyLimit = 2

    return self
end

function MultiplayerVehicleKeysManager:loadMap(filename)
    g_currentMission:registerObjectToCallOnMissionStart(self)

    g_messageCenter:subscribe(MessageType.PLAYER_NICKNAME_CHANGED, self.onPlayerNicknameChanged, self)
    g_messageCenter:subscribe(MessageType.PLAYER_FARM_CHANGED, self.onPlayerFarmChanged, self)

    if self.isServer then
        g_messageCenter:subscribe(MessageType.USER_REMOVED, self.onUserRemoved, self)

        local missionInfo = g_currentMission.missionInfo

        if missionInfo ~= nil and missionInfo.savegameDirectory ~= nil then
            local xmlFilename = missionInfo.savegameDirectory .. "/multiplayerVehicleKeys.xml"

            if fileExists(xmlFilename) then
                local xmlFile = loadXMLFile("MultiplayerVehicleKeysXML", xmlFilename)

                if xmlFile ~= nil and xmlFile ~= 0 then
                    self:setKeyLimit(getXMLInt(xmlFile, "multiplayerVehicleKeys.keyLimit"), true, true)

                    delete(xmlFile)
                end
            end
        end
    end

    -- Unfortunately this is needed as 'getCanStartAIVehicle' is not checked when using input binding
    AIJobVehicle.actionEventToggleAIState = Utils.overwrittenFunction(AIJobVehicle.actionEventToggleAIState, function(vehicle, superFunc, ...)
        if vehicle.getHasKeyOwner ~= nil and vehicle:getHasKeyOwner() and not vehicle:getIsKeyOwner() then
            local invalidUser = true

            if vehicle.isServer or g_currentMission.isMasterUser then
                invalidUser = not vehicle:getIsAIActive() -- Allow an admin or host to disable the AI but not start them
            end

            if invalidUser then
                local spec = vehicle.spec_enterableMultiplayerKeys
                local nickname = spec.keyOwnerLastNickname or ""

                if nickname ~= "" then
                    g_currentMission:showBlinkingWarning(string.format(spec.invalidUserWarning, nickname), 2000)
                end

                return
            end
        end

        superFunc(vehicle, ...)
    end)
end

function MultiplayerVehicleKeysManager:deleteMap()
    g_messageCenter:unsubscribeAll(self)

    if g_globalMods ~= nil then
        g_globalMods.multiplayerVehicleKeys = nil
    end

    g_multiplayerVehicleKeys = nil
end

function MultiplayerVehicleKeysManager:onMissionStarted()
    local text = g_i18n:getText("info_holdingKeys", modName)
    local hud = g_currentMission.hud

    for _, vehicle in ipairs(g_currentMission.enterables) do
        if vehicle.getIsKeyOwner ~= nil and vehicle:getIsKeyOwner() then
            hud:addSideNotification(MultiplayerVehicleKeys.INGAME_NOTIFICATION, string.format(text, tostring(vehicle:getFullName())), 6000)
        end
    end
end

function MultiplayerVehicleKeysManager:restoreAllVehicleKeys(suppressInfo, noEventSend)
    MultiplayerVehicleKeysSettingsEvent.sendEvent(true, self.keyLimit, noEventSend)

    local numReturned = 0

    for _, vehicle in ipairs(g_currentMission.enterables) do
        if vehicle.getHasKeyOwner ~= nil and vehicle:getHasKeyOwner() then
            vehicle:removeKeyOwner(true, true)

            numReturned = numReturned + 1
        end
    end

    if numReturned > 0 and not suppressInfo then
        Logging.info("[MultiplayerVehicleKeys] %d vehicle keys have been returned by an admin!", numReturned)
    end
end

function MultiplayerVehicleKeysManager:setKeyLimit(keyLimit, suppressInfo, noEventSend)
    self.keyLimit = MathUtil.clamp(keyLimit or 2, 1, 8)

    MultiplayerVehicleKeysSettingsEvent.sendEvent(false, self.keyLimit, noEventSend)

    if not suppressInfo then
        Logging.info("MultiplayerVehicleKeys Setting 'keyLimit': %d", keyLimit)
    end
end

function MultiplayerVehicleKeysManager:onPlayerNicknameChanged(player)
    if g_currentMission == nil or player == nil then
        return
    end

    local user = g_currentMission.userManager:getUserByUserId(player.userId)

    if user ~= nil then
        local nickname = user:getNickname()
        local uniqueUserId = user:getUniqueUserId()

        for _, vehicle in ipairs(g_currentMission.enterables) do
            local spec = vehicle.spec_enterableMultiplayerKeys

            if spec ~= nil and spec.keyOwnerUniqueUserId == uniqueUserId then
                spec.keyOwnerLastNickname = nickname
            end
        end
    end
end

function MultiplayerVehicleKeysManager:onPlayerFarmChanged(player)
    if player == g_currentMission.player then
        for _, vehicle in ipairs(g_currentMission.enterables) do
            if vehicle.getIsKeyOwner ~= nil and vehicle:getIsKeyOwner() then
                vehicle:removeKeyOwner(true)
            end
        end
    end
end

function MultiplayerVehicleKeysManager:onUserRemoved(user)
    if self.isServer and g_currentMission ~= nil and user:getIsBlocked() then
        local uniqueUserId = user:getUniqueUserId()

        if uniqueUserId ~= nil then
            for _, vehicle in ipairs(g_currentMission.enterables) do
                if vehicle.getIsKeyOwner ~= nil and vehicle:getIsKeyOwner(uniqueUserId) then
                    vehicle:removeKeyOwner(true)
                end
            end
        end
    end
end

function MultiplayerVehicleKeysManager:getCurrentKeys()
    return self.currentVehicleKeys
end

function MultiplayerVehicleKeysManager:getKeyLimit()
    return self.keyLimit
end

local function isActive()
    return g_modIsLoaded[modName] and g_multiplayerVehicleKeys ~= nil
end

local function validateTypes(typeManager)
    if typeManager.typeName == "vehicle" and isActive() then
        local specializationName = string.format("%s.multiplayerVehicleKeys", modName)
        local specializationObject = g_specializationManager:getSpecializationObjectByName(specializationName)

        if specializationObject ~= nil then
            local numInserted = 0

            for typeName, typeEntry in pairs (typeManager:getTypes()) do
                if specializationObject.prerequisitesPresent(typeEntry.specializations) then
                    typeManager:addSpecialization(typeName, specializationName)

                    numInserted = numInserted + 1
                end
            end

            g_multiplayerVehicleKeys.numInserted = numInserted
        end

        if g_multiplayerVehicleKeys.numInserted > 0 then
            -- No point restoring this on delete as so many other mods now also add custom settings, normally okay on PC anyway
            InGameMenuGameSettingsFrame.onFrameOpen = Utils.overwrittenFunction(InGameMenuGameSettingsFrame.onFrameOpen, function (frame, superFunc, element)
                superFunc(frame, element)

                if frame.multiMaxKeys == nil then
                    if frame.multiFuelUsage ~= nil and isActive() then
                        local boxLayout = frame.boxLayout
                        local multiElement = frame.multiFuelUsage

                        local headerText = "Multiplayer Vehicle Keys"
                        local mod = g_modManager:getModByName(modName)

                        if mod ~= nil and (mod.title ~= nil and mod.title ~= "") then
                            headerText = mod.title
                        end

                        for i = 1, #boxLayout.elements do
                            local element = boxLayout.elements[i]

                            if element:isa(TextElement) then
                                local header = element:clone(boxLayout)

                                header:setText(headerText)
                                header:reloadFocusHandling(true)

                                break
                            end
                        end

                        local resetKeysButtonParent = frame.buttonPauseGame.parent:clone(boxLayout)
                        local resetKeysButton = resetKeysButtonParent.elements[1]

                        function resetKeysButton.onClickCallback(_, element)
                            if g_multiplayerVehicleKeys ~= nil then
                                g_gui:showYesNoDialog({
                                    text = g_i18n:getText("warning_returnAllKeys", modName),
                                    yesText = g_i18n:getText("button_confirm"),
                                    noText = g_i18n:getText("button_cancel"),
                                    callback = function(yes)
                                        if yes then
                                            g_multiplayerVehicleKeys:restoreAllVehicleKeys()
                                        end
                                    end
                                })
                            end
                        end

                        resetKeysButton:setImageUVs(nil, GuiUtils.getUVs({292, 821, 89, 89}))
                        resetKeysButton:setText(g_i18n:getText("button_reset"))

                        resetKeysButtonParent.elements[2]:setText(g_i18n:getText("setting_resetVehicleKeys", modName))

                        local helpText = multiElement.elements[6]:clone(resetKeysButtonParent)
                        helpText:setText(g_i18n:getText("toolTip_resetVehicleKeys", modName))

                        resetKeysButton.id = "resetKeysButton"
                        frame.resetKeysButton = resetKeysButton

                        resetKeysButtonParent:reloadFocusHandling(true)

                        local multiMaxKeysTexts = {}

                        for i = 1, 8 do
                            multiMaxKeysTexts[i] = tostring(i)
                        end

                        local multiMaxKeys = frame.multiFuelUsage:clone(boxLayout)

                        multiMaxKeys:setTexts(multiMaxKeysTexts)

                        function multiMaxKeys.onClickCallback(_, keyLimit)
                            if g_multiplayerVehicleKeys ~= nil then
                                g_multiplayerVehicleKeys:setKeyLimit(keyLimit, false)
                            end
                        end

                        multiMaxKeys.id = "multiMaxKeys"
                        frame.multiMaxKeys = multiMaxKeys

                        multiMaxKeys.elements[4]:setText(g_i18n:getText("setting_maxVehicleKeys", modName))
                        multiMaxKeys.elements[6]:setText(g_i18n:getText("toolTip_maxVehicleKeys", modName))

                        multiMaxKeys:reloadFocusHandling(true)
                        multiMaxKeys:setState(g_multiplayerVehicleKeys:getKeyLimit())

                        -- Fix linked elements so 'Pause Button' is targeted from bottom and last element from 'Pause Button'
                        local lastElement = boxLayout.elements[#boxLayout.elements]

                        FocusManager:linkElements(lastElement, FocusManager.BOTTOM, frame.buttonPauseGame)
                        FocusManager:linkElements(frame.buttonPauseGame, FocusManager.TOP, lastElement)

                        boxLayout:invalidateLayout()
                    end
                elseif g_multiplayerVehicleKeys ~= nil then
                    frame.multiMaxKeys:setState(g_multiplayerVehicleKeys:getKeyLimit())
                end
            end)
        end
    end
end

local function saveToXMLFile(missionInfo)
    if missionInfo.isValid and isActive() then
        local xmlFilename = missionInfo.savegameDirectory .. "/multiplayerVehicleKeys.xml"
        local xmlFile = createXMLFile("MultiplayerVehicleKeysXML", xmlFilename, "multiplayerVehicleKeys")

        if xmlFile ~= nil and xmlFile ~= 0 then
            setXMLString(xmlFile, "multiplayerVehicleKeys#version", versionString)
            setXMLFloat(xmlFile, "multiplayerVehicleKeys#buildId", buildId)

            setXMLInt(xmlFile, "multiplayerVehicleKeys.keyLimit", g_multiplayerVehicleKeys.keyLimit or 2)

            saveXMLFile(xmlFile)
            delete(xmlFile)
        end
    end
end

local function onConnectionFinishedLoading(mission, connection, x, y, z, viewDistanceCoeff)
    if mission:getIsServer() and isActive() then
        connection:sendEvent(MultiplayerVehicleKeysSettingsEvent.new(false, g_multiplayerVehicleKeys:getKeyLimit()))
    end
end

local function getIsMultiplayerGame()
    if buildId > 0 then
        if g_dedicatedServerInfo ~= nil then
            return true
        end

        if g_careerScreen ~= nil then
            return g_careerScreen.isMultiplayer
        end
    end

    return true
end

local function init()
    if g_globalMods and g_globalMods.multiplayerVehicleKeys then
        Logging.error("Failed to load mod '%s', 'MultiplayerVehicleKeys' specialisation was already loaded by '%s'!", modName, g_globalMods.multiplayerVehicleKeys.modName or "N/A")

        return
    end

    local isValidMod = false
    local mod = g_modManager:getModByName(modName)

    versionString = mod.version or versionString

    if mod.modName == "FS22_MultiplayerVehicleKeys" or mod.modName == "FS22_MultiplayerVehicleKeys_update" then
        isValidMod = mod.author and #mod.author == 3
    end

    if isValidMod then
        if getIsMultiplayerGame() then
            source(modDirectory .. "scripts/events/MultiplayerVehicleKeysSettingsEvent.lua")
            source(modDirectory .. "scripts/events/MultiplayerVehicleKeysSetOwnerEvent.lua")

            g_specializationManager:addSpecialization("multiplayerVehicleKeys", "MultiplayerVehicleKeys", modDirectory .. "scripts/specializations/MultiplayerVehicleKeys.lua", nil)

            local multiplayerVehicleKeys = MultiplayerVehicleKeysManager.new()

            multiplayerVehicleKeys.version = version
            multiplayerVehicleKeys.buildId = buildId

            g_multiplayerVehicleKeys = multiplayerVehicleKeys
            g_globalMods.multiplayerVehicleKeys = multiplayerVehicleKeys

            addModEventListener(multiplayerVehicleKeys)

            TypeManager.validateTypes = Utils.prependedFunction(TypeManager.validateTypes, validateTypes)
            FSCareerMissionInfo.saveToXMLFile = Utils.appendedFunction(FSCareerMissionInfo.saveToXMLFile, saveToXMLFile)
            FSBaseMission.onConnectionFinishedLoading = Utils.appendedFunction(FSBaseMission.onConnectionFinishedLoading, onConnectionFinishedLoading)
        else
            popupMessage = {
                startUpdateTime = 2000,

                update = function(self, dt)
                    self.startUpdateTime = self.startUpdateTime - dt

                    if self.startUpdateTime < 0 and not g_gui:getIsGuiVisible() then
                        if g_currentMission.hud ~= nil then
                            local message = string.format(g_i18n:getText("warning_multiplayerNotDetected", modName), modName)

                            g_currentMission.hud:showInGameMessage("", message or "", -1, nil, nil, nil)
                        end

                        removeModEventListener(self)
                        popupMessage = nil
                    end
                end
            }

            addModEventListener(popupMessage)
        end
    else
        validationFail = {
            startUpdateTime = 2000,

            update = function(self, dt)
                self.startUpdateTime = self.startUpdateTime - dt

                if self.startUpdateTime < 0 then
                    local text = string.format(g_i18n:getText("mvk_loadError", modName), modName, mod.author or "Unknown")

                    if g_dedicatedServerInfo == nil then
                        if not g_gui:getIsGuiVisible() then
                            g_gui:showYesNoDialog({
                                title = string.format("%s - Version %s", mod.title, versionString),
                                text = text,
                                dialogType = DialogElement.TYPE_LOADING,
                                callback = self.openModHubLink,
                                yesText = g_i18n:getText("button_ok"),
                                noText = g_i18n:getText("button_modHubDownload")
                            })
                        end
                    else
                        print("\n" .. text .. "\n    - https://farming-simulator.com/mods.php?lang=en&country=be&title=fs2022&filter=org&org_id=129652&page=0" .. "\n")
                        self.openModHubLink(true)
                    end
                end
            end,

            openModHubLink = function(ignore)
                if ignore == false then
                    local language = g_languageShort
                    local link = "mods.php?lang=en&country=be&title=fs2022&filter=org&org_id=129652&page=0"
                    if language == "de" or language == "fr" then
                        link = "mods.php?lang=" .. language .. "&country=be&title=fs2022&filter=org&org_id=129652&page=0"
                    end

                    openWebFile(link, "")
                end

                removeModEventListener(validationFail)
                validationFail = nil
            end
        }

        addModEventListener(validationFail)
    end
end

init()
